from osm2gmns.networkclass.macronet import Node, Link, Movement, Segment, POI, Network
from osm2gmns.movement.util_mvmt import getMovementDescription, getMovementGeometry
from osm2gmns.utils.util_coord import from_latlon
from osm2gmns.utils.util_geo import GeoTransformer
import osm2gmns.settings as og_settings
import os
import csv
import sys
import numpy as np
from shapely import wkt, geometry


_node_required_fields = {'node_id', 'x_coord', 'y_coord'}
_node_optional_fields = {'name', 'osm_node_id', 'osm_highway', 'zone_id', 'ctrl_type', 'node_type', 'activity_type', 'is_boundary', 'intersection_id', 'poi_id', 'notes'}
_link_required_fields = {'link_id', 'from_node_id', 'to_node_id', 'lanes'}
_link_optional_fields = {'name', 'osm_way_id', 'dir_flag', 'length', 'free_speed', 'capacity', 'link_type_name', 'link_type', 'geometry', 'allowed_uses', 'from_biway', 'is_link', 'VDF_fftt1', 'VDF_cap1'}
_mvmt_required_fields = {'mvmt_id', 'node_id', 'ib_link_id', 'start_ib_lane', 'end_ib_lane', 'ob_link_id', 'start_ob_lane', 'end_ob_lane'}
_mvmt_optional_fields = {'osm_node_id', 'name', 'lanes', 'ib_osm_node_id', 'ob_osm_node_id', 'type', 'penalty', 'capacity', 'ctrl_type', 'mvmt_txt_id', 'geometry', 'volume', 'free_speed', 'allowed_uses','generated_by_osm2gmns'}
_segment_required_fields = {'segment_id', 'link_id', 'ref_node_id', 'start_lr', 'end_lr', 'l_lanes_added', 'r_lanes_added'}
_segment_optional_fields = {'grade', 'capacity', 'free_speed', 'lanes', 'bike_facility', 'ped_facility', 'parking', 'allowed_uses', 'toll', 'jurisdiction', 'row_width', 'opt_comment'}


def _loadNodes(network, node_filepath, coordinate_type, encoding):
    if encoding is None:
        fin = open(node_filepath, 'r')
    else:
        fin = open(node_filepath, 'r', encoding=encoding)
    reader = csv.DictReader(fin)

    for field in _node_required_fields:
        if field not in reader.fieldnames:
            sys.exit(f'ERROR: required field ({field}) does not exist in the node file')
    other_fields = list(set(reader.fieldnames).difference(_node_required_fields.union(_node_optional_fields)))

    max_node_id = network.max_node_id
    max_intersection_id = network.max_intersection_id
    node_id_list = []
    node_coord_list = []
    node_dict = {}
    for node_info in reader:
        # required
        node = Node(int(node_info['node_id']))
        if node.node_id > max_node_id: max_node_id = node.node_id

        x_coord, y_coord = float(node_info['x_coord']), float(node_info['y_coord'])
        if coordinate_type == 'lonlat':
            node.geometry = geometry.Point(round(x_coord, og_settings.lonlat_coord_precision), round(y_coord, og_settings.lonlat_coord_precision))
            node_id_list.append(node.node_id)
            node_coord_list.append((x_coord, y_coord))
        elif coordinate_type == 'meter':
            node.geometry_xy = geometry.Point(round(x_coord, og_settings.local_coord_precision), round(y_coord, og_settings.local_coord_precision))
        elif coordinate_type == 'feet':
            node.geometry_xy = geometry.Point(round(x_coord*0.3048, og_settings.local_coord_precision), round(y_coord*0.3048, og_settings.local_coord_precision))

        # optional
        name = node_info['name'] if 'name' in reader.fieldnames else None
        if name: node.name = name
        osm_node_id = node_info['osm_node_id'] if 'osm_node_id' in reader.fieldnames else None
        if osm_node_id: node.osm_node_id = osm_node_id
        osm_highway = node_info['osm_highway'] if 'osm_highway' in reader.fieldnames else None
        if osm_highway: node.osm_highway = osm_highway
        zone_id = node_info['zone_id'] if 'zone_id' in reader.fieldnames else None
        if zone_id: node.zone_id = int(zone_id)

        ctrl_type = node_info['ctrl_type'] if 'ctrl_type' in reader.fieldnames else None
        if ctrl_type: node.ctrl_type = ctrl_type
        node_type = node_info['node_type'] if 'node_type' in reader.fieldnames else None
        if node_type: node.node_type = node_type
        activity_type = node_info['activity_type'] if 'activity_type' in reader.fieldnames else None
        if activity_type: node.activity_type = activity_type
        is_boundary = node_info['is_boundary'] if 'is_boundary' in reader.fieldnames else None
        if is_boundary: node.is_boundary = int(is_boundary)

        intersection_id = node_info['intersection_id'] if 'intersection_id' in reader.fieldnames else None
        if intersection_id:
            node.intersection_id = int(intersection_id)
            if node.intersection_id > max_intersection_id:
                max_intersection_id = node.intersection_id

        poi_id = node_info['poi_id'] if 'poi_id' in reader.fieldnames else None
        if poi_id: node.poi_id = poi_id
        notes = node_info['notes'] if 'notes' in reader.fieldnames else None
        if notes: node.notes = notes

        # others
        for field in other_fields:
            node.other_attrs[field] = node_info[field]

        node_dict[node.node_id] = node
    fin.close()

    max_node_id += 1
    max_intersection_id += 1

    if coordinate_type == 'lonlat':
        coord_array = np.array(node_coord_list)
        central_lon = coord_array[:,0].mean()
        central_lat = coord_array[:,1].mean()
        northern = True if central_lat >= 0 else False

        xs, ys = from_latlon(coord_array[:,0], coord_array[:,1], central_lon)
        xs, ys = np.round(xs, og_settings.local_coord_precision), np.round(ys, og_settings.local_coord_precision)

        for node_no, node_id in enumerate(node_id_list):
            node = node_dict[node_id]
            node.geometry_xy = geometry.Point((xs[node_no],ys[node_no]))

        network.GT = GeoTransformer(central_lon, central_lat, northern)

    network.node_dict = node_dict
    network.max_node_id = max_node_id
    network.max_intersection_id = max_intersection_id
    network.node_other_attrs = other_fields


def _loadLinks(network, link_filepath, coordinate_type, encoding):
    if encoding is None:
        fin = open(link_filepath, 'r')
    else:
        fin = open(link_filepath, 'r', encoding=encoding)
    reader = csv.DictReader(fin)

    for field in _link_required_fields:
        if field not in reader.fieldnames:
            sys.exit(f'ERROR: required field ({field}) does not exist in the link file')
    other_fields = list(set(reader.fieldnames).difference(_link_required_fields.union(_link_optional_fields)))

    node_dict = network.node_dict
    GT = network.GT

    max_link_id = network.max_link_id
    link_dict = {}
    for link_info in reader:
        # required
        link = Link(int(link_info['link_id']))
        if link.link_id > max_link_id: max_link_id = link.link_id

        from_node_id, to_node_id = int(link_info['from_node_id']), int(link_info['to_node_id'])
        try:
            link.from_node = node_dict[from_node_id]
        except KeyError:
            print(f'WARNING: from_node {from_node_id} of link {link.link_id} does not exist in the node file')
            continue
        try:
            link.to_node = node_dict[to_node_id]
        except KeyError:
            print(f'WARNING: to_node {to_node_id} of link {link.link_id} does not exist in the node file')
            continue

        lanes = link_info['lanes']
        if lanes:
            link.lanes_list = [int(lanes)]
        else:
            link.lanes_list = [None]

        # optional
        name = link_info['name'] if 'name' in reader.fieldnames else None
        if name: link.name = name
        osm_way_id = link_info['osm_way_id'] if 'osm_way_id' in reader.fieldnames else None
        if osm_way_id: link.osm_way_id = osm_way_id
        dir_flag = link_info['dir_flag'] if 'dir_flag' in reader.fieldnames else None
        if dir_flag: link.dir_flag = int(dir_flag)

        free_speed = link_info['free_speed'] if 'free_speed' in reader.fieldnames else None
        if free_speed: link.free_speed = float(free_speed)
        capacity = link_info['capacity'] if 'capacity' in reader.fieldnames else None
        if capacity: link.capacity = float(capacity)

        link_type_name = link_info['link_type_name'] if 'link_type_name' in reader.fieldnames else None
        if link_type_name: link.link_type_name = link_type_name
        link_type = link_info['link_type'] if 'link_type' in reader.fieldnames else None
        if link_type:
            link.link_type = int(link_type)

        geometry_str = link_info['geometry'] if 'geometry' in reader.fieldnames else None
        if geometry_str:
            if coordinate_type == 'lonlat':
                link.geometry = wkt.loads(geometry_str)
                link.geometry_xy = GT.geo_from_latlon(link.geometry)
            elif coordinate_type == 'meter':        # todo
                pass
            elif coordinate_type == 'feet':         # todo
                pass

        allowed_uses = link_info['allowed_uses'] if 'allowed_uses' in reader.fieldnames else None
        if allowed_uses: link.allowed_uses = allowed_uses.split(';')

        from_biway = link_info['from_biway'] if 'from_biway' in reader.fieldnames else None
        if from_biway:
            if from_biway == '1':
                link.from_bidirectional_way = True
            elif from_biway == '0':
                link.from_bidirectional_way = False

        is_link = link_info['is_link'] if 'is_link' in reader.fieldnames else None
        if is_link:
            if is_link == '1':
                link.is_link = True
            elif is_link == '0':
                link.is_link = False

        VDF_fftt1 = link_info['VDF_fftt1'] if 'VDF_fftt1' in reader.fieldnames else None
        if VDF_fftt1: link.VDF_fftt1 = float(VDF_fftt1)
        VDF_cap1 = link_info['VDF_cap1'] if 'VDF_cap1' in reader.fieldnames else None
        if VDF_cap1: link.VDF_cap1 = float(VDF_cap1)

        # others
        for field in other_fields:
            link.other_attrs[field] = link_info[field]

        link_dict[link.link_id] = link
        link.from_node.outgoing_link_list.append(link)
        link.to_node.incoming_link_list.append(link)

    fin.close()
    max_link_id += 1

    network.max_link_id = max_link_id
    network.link_dict = link_dict
    network.link_other_attrs = other_fields


def _loadMovements(network, movement_filepath, coordinate_type, encoding):
    if encoding is None:
        fin = open(movement_filepath, 'r')
    else:
        fin = open(movement_filepath, 'r', encoding=encoding)
    reader = csv.DictReader(fin)

    for field in _mvmt_required_fields:
        if field not in reader.fieldnames:
            sys.exit(f'ERROR: required field ({field}) does not exist in the movement file')
    other_fields = list(set(reader.fieldnames).difference(_mvmt_required_fields.union(_mvmt_optional_fields)))

    node_dict = network.node_dict
    link_dict = network.link_dict
    GT = network.GT

    max_movement_id = network.max_movement_id

    for movement_info in reader:
        # required
        movement = Movement(int(movement_info['mvmt_id']))
        if movement.movement_id > max_movement_id: max_movement_id = movement.movement_id

        node_id = int(movement_info['node_id'])
        try:
            node = node_dict[node_id]
        except KeyError:
            print(f'WARNING: node {node_id} of movement {movement.movement_id} does not exist in the node file')
            continue

        ib_link_id = int(movement_info['ib_link_id'])
        try:
            ib_link = link_dict[ib_link_id]
        except KeyError:
            print(f'WARNING: ib_link {ib_link_id} of movement {movement.movement_id} does not exist in the link file')
            continue
        if ib_link.lanes is None:
            print(f'WARNING: ib_link {ib_link_id} of movement {movement.movement_id} does not have lanes information')
            continue
        start_ib_lane = int(movement_info['start_ib_lane'])
        if start_ib_lane > ib_link.lanes:
            print(f'WARNING: start_ib_lane of movement {movement.movement_id} is larger than total lanes of ib_link {ib_link_id}')
            continue
        end_ib_lane_str = movement_info['end_ib_lane']
        end_ib_lane = int(end_ib_lane_str) if end_ib_lane_str else start_ib_lane

        ob_link_id = int(movement_info['ob_link_id'])
        try:
            ob_link = link_dict[ob_link_id]
        except KeyError:
            print(f'WARNING: ob_link {ob_link_id} of movement {movement.movement_id} does not exist in the link file')
            continue
        if ob_link.lanes is None:
            print(f'WARNING: ob_link {ob_link_id} of movement {movement.movement_id} does not have lanes information')
            continue
        start_ob_lane = int(movement_info['start_ob_lane'])
        if start_ob_lane > ob_link.lanes:
            print(f'WARNING: start_ob_lane of movement {movement.movement_id} is larger than total lanes of ob_link {ob_link_id}')
            continue
        end_ob_lane_str = movement_info['end_ob_lane']
        end_ob_lane = int(end_ob_lane_str) if end_ob_lane_str else start_ob_lane

        ib_lanes, ob_lanes = end_ib_lane - start_ib_lane + 1, end_ob_lane - start_ob_lane + 1
        if ib_lanes != ob_lanes:
            print(f'WARNING: ib_lanes ({ib_lanes}) and ob_lanes ({ob_lanes}) of movement {movement.movement_id} are not consistent')
            continue
        movement.lanes = ib_lanes

        movement.ib_link, movement.start_ib_lane, movement.end_ib_lane = ib_link, start_ib_lane, end_ib_lane
        movement.ob_link, movement.start_ob_lane, movement.end_ob_lane = ob_link, start_ob_lane, end_ob_lane

        # optional
        type_file = movement_info['type'] if 'type' in reader.fieldnames else None
        mvmt_txt_id_file = movement_info['mvmt_txt_id'] if 'mvmt_txt_id' in reader.fieldnames else None
        if type_file and mvmt_txt_id_file:
            movement.type, movement.mvmt_txt_id = type_file, mvmt_txt_id_file
        else:
            mvmt_txt_id, mvmt_type = getMovementDescription(ib_link, ob_link)
            movement.type = type_file if type_file else mvmt_type
            movement.mvmt_txt_id = mvmt_txt_id_file if mvmt_txt_id_file else mvmt_txt_id

        penalty = movement_info['penalty'] if 'penalty' in reader.fieldnames else None
        if penalty: movement.penalty = float(penalty)
        capacity = movement_info['capacity'] if 'capacity' in reader.fieldnames else None
        if capacity: movement.capacity = float(capacity)
        ctrl_type = movement_info['ctrl_type'] if 'ctrl_type' in reader.fieldnames else None
        if ctrl_type: movement.ctrl_type = ctrl_type
        volume = movement_info['volume'] if 'volume' in reader.fieldnames else None
        if volume: movement.volume = float(volume)
        free_speed = movement_info['free_speed'] if 'free_speed' in reader.fieldnames else None
        if free_speed: movement.free_speed = float(free_speed)
        allowed_uses = movement_info['allowed_uses'] if 'allowed_uses' in reader.fieldnames else None
        if allowed_uses: movement.allowed_uses = allowed_uses.split(';')
        generated_by_osm2gmns = movement_info['generated_by_osm2gmns'] if 'generated_by_osm2gmns' in reader.fieldnames else None
        if generated_by_osm2gmns:
            if generated_by_osm2gmns == '1':
                movement.generated_by_osm2gmns = True
            elif generated_by_osm2gmns == '0':
                movement.generated_by_osm2gmns = False

        geometry_str = movement_info['geometry'] if 'geometry' in reader.fieldnames else None
        if geometry_str:
            if coordinate_type == 'lonlat':
                movement.geometry = wkt.loads(geometry_str)
                movement.geometry_xy = GT.geo_from_latlon(movement.geometry)
            elif coordinate_type == 'meter':        # todo
                pass
            elif coordinate_type == 'feet':         # todo
                pass
        else:
            if coordinate_type == 'lonlat':
                movement.geometry_xy = getMovementGeometry(ib_link, ob_link)
                movement.geometry = GT.geo_to_latlon(movement.geometry_xy)
            elif coordinate_type == 'meter':        # todo
                pass
            elif coordinate_type == 'feet':         # todo
                pass

        # others
        for field in other_fields:
            movement.other_attrs[field] = movement_info[field]

        node.movement_list.append(movement)

    fin.close()
    max_movement_id += 1

    network.max_movement_id = max_movement_id
    network.movement_other_attrs = other_fields


def _loadSegments(network, segment_filepath, encoding):
    if encoding is None:
        fin = open(segment_filepath, 'r')
    else:
        fin = open(segment_filepath, 'r', encoding=encoding)
    reader = csv.DictReader(fin)

    for field in _segment_required_fields:
        if field not in reader.fieldnames:
            sys.exit(f'ERROR: required field ({field}) does not exist in the segment file')
    other_fields = list(set(reader.fieldnames).difference(_segment_required_fields.union(_segment_optional_fields)))

    node_dict = network.node_dict
    link_dict = network.link_dict

    max_segment_id = network.max_segment_id

    for segment_info in reader:
        # required
        segment = Segment(int(segment_info['segment_id']))
        if segment.segment_id > max_segment_id: max_segment_id = segment.segment_id

        link_id = int(segment_info['link_id'])
        try:
            link = link_dict[link_id]
        except KeyError:
            print(f'WARNING: link {link_id} of segment {segment.segment_id} does not exist in the link file')
            continue

        if link.lanes is None:
            print(f'WARNING: link {link_id} of segment {segment.segment_id} does not have lanes information')
            continue

        segment.link = link

        ref_node_id = int(segment_info['ref_node_id'])
        try:
            ref_node = node_dict[ref_node_id]
        except KeyError:
            print(f'WARNING: ref_node {ref_node_id} of segment {segment.segment_id} does not exist in the node file')
            continue
        segment.ref_node = ref_node

        segment.start_lr = float(segment_info['start_lr'])
        segment.end_lr = float(segment_info['end_lr'])
        segment.l_lanes_added = int(segment_info['l_lanes_added'])
        segment.r_lanes_added = int(segment_info['r_lanes_added'])

        # optional

        # others
        for field in other_fields:
            segment.other_attrs[field] = segment_info[field]

        link.segment_list.append(segment)

    fin.close()
    max_segment_id += 1

    network.max_segment_id = max_segment_id
    network.segment_other_attrs = other_fields


def _loadGeometries(network, geometry_filepath, encoding):
    pass


def _loadPOIs(network, poi_filepath, encoding):
    if encoding is None:
        fin = open(poi_filepath, 'r')
    else:
        fin = open(poi_filepath, 'r', encoding=encoding)
    reader = csv.DictReader(fin)

    max_poi_id = network.max_poi_id
    GT = network.GT

    POI_list = []
    for poi_info in reader:
        poi = POI()
        name = poi_info['name']
        if name: poi.name = name
        poi.poi_id = int(poi_info['poi_id'])
        if poi.poi_id > max_poi_id: max_poi_id = poi.poi_id
        osm_way_id = poi_info['osm_way_id']
        if osm_way_id: poi.osm_way_id = osm_way_id
        osm_relation_id = poi_info['osm_relation_id']
        if osm_relation_id: poi.osm_relation_id = osm_relation_id
        building = poi_info['building']
        if building: poi.building = building
        amenity = poi_info['amenity']
        if amenity: poi.amenity = amenity
        leisure = poi_info['leisure']
        if leisure: poi.leisure = leisure
        way = poi_info['way']
        if way: poi.way = way

        poi.geometry = wkt.loads(poi_info['geometry'])
        lon, lat = poi.geometry.centroid.x, poi.geometry.centroid.y
        poi.centroid = geometry.Point((round(lon,og_settings.lonlat_coord_precision),round(lat,og_settings.lonlat_coord_precision)))

        poi.geometry_xy = GT.geo_from_latlon(poi.geometry)
        x, y = poi.geometry_xy.centroid.x, poi.geometry_xy.centroid.y
        poi.centroid_xy = geometry.Point((round(x,og_settings.local_coord_precision),round(y,og_settings.local_coord_precision)))

        POI_list.append(poi)

    fin.close()
    max_poi_id += 1

    network.max_poi_id = max_poi_id



def getNetFromCSV(folder='', enconding=None):
    """
    Deprecated
    """
    print('Warning: getNetFromCSV() is deprecated and will be removed in a future release. Please use loadNetFromCSV().\n'
          '         For more information, please refer to the document at https://osm2gmns.readthedocs.io')
    network = loadNetFromCSV(folder, enconding)
    return network


def loadNetFromCSV(folder='', node_file=None, link_file=None, movement_file=None,
                   segment_file=None, geometry_file=None, POI_file=None,
                   coordinate_type='lonlat', enconding=None):
    """
    Load a network from csv files in GMNS format

    Parameters
    ----------
    folder: str
        the folder where network files are stored
    node_file: str
        filename of the node file. required
    link_file: str
        filename of the link file. required
    movement_file: str, None
        filename of the movement file. optional
    segment_file: str, None
        filename of the segment file. optional
    geometry_file: str, None
        filename of the geometry file. optional
    POI_file: str, None
        filename of the POI file. optional
    coordinate_type: str
        the coordinate system used by the network to be loaded. can be lonlat, meter, feet
    enconding: str, None
        the encoding used by the network files. if None, osm2gmns will use the default encoding of the local operating system

    Returns
    -------
    network: Network
        an osm2gmns Network object

    """

    if node_file is None:
        sys.exit('ERROR: node_file is not specified')
    node_filepath = os.path.join(folder, node_file)
    if not os.path.isfile(node_filepath):
        sys.exit(f'ERROR: cannot open node_file {node_filepath}')

    if link_file is None:
        sys.exit('ERROR: link_file is not specified')
    link_filepath = os.path.join(folder, link_file)
    if not os.path.isfile(link_filepath):
        sys.exit(f'ERROR: cannot open link_file {link_filepath}')

    if movement_file is not None:
        movement_filepath = os.path.join(folder, movement_file)
        if not os.path.isfile(movement_filepath):
            sys.exit(f'ERROR: cannot open movement_file {movement_filepath}')
    else:
        movement_filepath = None

    if segment_file is not None:
        segment_filepath = os.path.join(folder, segment_file)
        if not os.path.isfile(segment_filepath):
            sys.exit(f'ERROR: cannot open segment_file {segment_filepath}')
    else:
        segment_filepath = None

    if geometry_file is not None:
        geometry_filepath = os.path.join(folder, geometry_file)
        if not os.path.isfile(geometry_filepath):
            sys.exit(f'ERROR: cannot open geometry_file {geometry_filepath}')
    else:
        geometry_filepath = None

    if POI_file is not None:
        POI_filepath = os.path.join(folder, POI_file)
        if not os.path.isfile(POI_filepath):
            sys.exit(f'ERROR: cannot open POI_file {POI_filepath}')
    else:
        POI_filepath = None

    network = Network()

    _loadNodes(network, node_filepath, coordinate_type, enconding)

    _loadLinks(network, link_filepath, coordinate_type, enconding)

    if movement_file is not None:
        _loadMovements(network, movement_filepath, coordinate_type, enconding)

    if segment_file is not None:
        _loadSegments(network, segment_filepath, enconding)

    if geometry_file is not None:
        _loadGeometries(network, geometry_filepath, enconding)

    if POI_file is not None:
        _loadPOIs(network, POI_filepath, enconding)

    return network


# @author       Jiawei Lu (jiaweil9@asu.edu)
# @time         2021/11/27 15:34
# @desc         [script description]

from osm2gmns.networkclass.basenet import BaseNode, BaseLink, BaseNetwork
from osm2gmns.osmnet.osmclasses import OSMNode
from osm2gmns.utils.util_geo import getLineFromNodes
import osm2gmns.settings as og_settings
import math


class Node(BaseNode):
    def __init__(self, node_id):
        super().__init__(node_id)
        self.zone_id = None
        self.osm_node_id = None     # str
        self.intersection_id = None
        self.osm_highway = None
        self.node_type = ''
        self.ctrl_type = ''
        self.activity_type = ''
        self.controller_id = None
        self.is_boundary = None        # -1:in, 1:out, 2:in&out, 0:no
        self.poi_id = None
        self.notes = ''

        self.movement_list = []

        # ======== FOR MULTIRESOLUTION ========
        self.is_centroid = False
        self.centroid_mesonode = None
        self.movement_link_needed = True


    def buildFromOSMNode(self, osmnode):
        """
        initialize the node from an osmnode

        Parameters
        ----------
        osmnode: OSMNode
            osmnode instance

        Returns
        -------

        """
        self.name = osmnode.name
        self.osm_node_id = osmnode.osm_node_id
        self.osm_highway = osmnode.osm_highway
        self.ctrl_type = osmnode.ctrl_type
        self.geometry = osmnode.geometry
        self.geometry_xy = osmnode.geometry_xy
        self.notes = osmnode.notes



class Link(BaseLink):
    def __init__(self, link_id):
        super().__init__(link_id)
        self.osm_way_id = None      # str
        self.free_speed = None
        self.capacity = None
        self.link_class = ''        # highway, railway, aeroway
        self.link_type_name = ''
        self.link_type = 0
        self.is_link = False
        self.is_connector = False

        self.lanes_list = [None]
        self.from_bidirectional_way = False
        self.VDF_fftt1 = None
        self.VDF_cap1 = None
        self.ctrl_type = None       # signal node in the middle

        self.segment_list = []

        # ======== FOR MULTIRESOLUTION ========
        self.lanes_change_point_list = []
        self.lanes_change_list = []
        self.geometry_offset = None
        self.geometry_xy_offset = None

        self.upstream_short_cut = False         # no movement needed at upstream
        self.upstream_is_target = False         # connect with other links while keeping its upstream unchanged
        self.downstream_short_cut = False       # no movement needed at downstream
        self.downstream_is_target = False       # connect with other links while keeping its upstream unchanged

        self.length_of_cut_upstream = 0.0
        self.length_of_cut_downstream = 0.0

        # self.cutted_number_of_sections = 0
        # self.cutted_length = 0.0
        # self.cutted_length_in_km = 0.0
        self.cutted_geometry_list = []      # for each segment
        self.cutted_geometry_xy_list = []  # for each segment
        self.cutted_lanes_list = []
        self.cutted_lanes_change_point_list = []
        self.cutted_lanes_change_list = []

        self.mesolink_list = []


    @property
    def lanes(self):
        return self.lanes_list[0]
    @property
    def outgoing_lanes(self):
        return self.lanes_list[-1]
    @property
    def max_lanes(self):
        return None if None in self.lanes_list else max(self.lanes_list)
    @property
    def length_offset(self):
        return round(self.geometry_xy_offset.length, og_settings.local_coord_precision)


    def buildFromOSMWay(self, way, direction, ref_node_list, default_lanes, default_speed, default_capacity):
        self.osm_way_id = way.osm_way_id
        self.name = way.name
        self.link_class = way.link_class
        self.link_type_name = way.link_type_name
        self.link_type = way.link_type
        self.is_link = way.is_link

        if way.maxspeed:
            self.free_speed = way.maxspeed
        elif default_speed:
            self.free_speed = default_speed[self.link_type_name]

        if default_capacity:
            self.capacity = default_capacity[self.link_type_name]

        self.allowed_uses = way.allowed_uses
        if not way.oneway: self.from_bidirectional_way = True

        if way.oneway:
            self.lanes_list = [way.lanes]
        else:
            if direction == 1:
                if way.forward_lanes is not None:
                    self.lanes_list = [way.forward_lanes]
                elif way.lanes is not None:
                    self.lanes_list = [math.ceil(way.lanes / 2)]
                else:
                    self.lanes_list = [way.lanes]
            else:
                if way.backward_lanes is not None:
                    self.lanes_list = [way.backward_lanes]
                elif way.lanes is not None:
                    self.lanes_list = [math.ceil(way.lanes / 2)]
                else:
                    self.lanes_list = [way.lanes]

        if (self.lanes is None) and default_lanes:
            self.lanes_list = [default_lanes[self.link_type_name]]

        for ref_node in ref_node_list[1:-1]:
            if ref_node.ctrl_type == 'signal':
                self.ctrl_type = 'signal'

        self.from_node = ref_node_list[0].node
        self.to_node = ref_node_list[-1].node
        self.geometry, self.geometry_xy = getLineFromNodes(ref_node_list)
        self.from_node.outgoing_link_list.append(self)
        self.to_node.incoming_link_list.append(self)


class Segment:
    def __init__(self, segment_id):
        self.segment_id = segment_id
        self.link = None
        self.ref_node = None
        self.start_lr = 0.0
        self.end_lr = 0.0
        self.l_lanes_added = 0
        self.r_lanes_added = 0

        self.other_attrs = {}


class Movement:
    def __init__(self, movement_id):
        self.movement_id = movement_id
        self.node = None
        self.ib_link = None
        self.ob_link = None

        self.start_ib_lane = None
        self.end_ib_lane = None
        self.start_ob_lane = None
        self.end_ob_lane = None

        self.name = ''
        self.lanes = 0
        self.type = ''
        self.penalty = 0.0
        self.capacity = 0.0
        self.ctrl_type = ''
        self.volume = None
        self.free_speed = None
        # self.intersection_id = None
        self.mvmt_txt_id = ''
        self.allowed_uses = []

        self.geometry = None
        self.geometry_xy = None

        self.generated_by_osm2gmns = False

        self.other_attrs = {}


class POI:
    def __init__(self):
        self.poi_id = 0
        self.osm_way_id = None      # str
        self.osm_relation_id = None
        self.name = None
        self.geometry = None
        self.geometry_xy = None
        self.centroid = None
        self.centroid_xy = None
        self.nearest_node = None
        self.building = None
        self.amenity = None
        self.leisure = None
        self.way = None         # highway,railway,aeroway poi


class Network(BaseNetwork):
    def __init__(self):
        super().__init__()

        self.default_lanes = False
        self.default_speed = False
        self.default_capacity = False

        self.max_intersection_id = 0
        self.max_segment_id = 0
        self.max_poi_id = 0
        self.max_movement_id = 0

        self.POI_list = []

        self.movement_other_attrs = []
        self.segment_other_attrs = []

        self.mesonet = None
        self.micronet = None


    @property
    def number_of_pois(self):
        return len(self.POI_list)




from osm2gmns.osmnet.osmclasses import OSMNode, Way, Relation, OSMNetwork
from osm2gmns.utils.util_coord import from_latlon
from osm2gmns.utils.util_geo import GeoTransformer
from osm2gmns.utils.util import getLogger
import osm2gmns.settings as og_settings
from shapely import geometry
import numpy as np
import sys
import osmium
import re


class NWRHandler(osmium.SimpleHandler):
    def __init__(self):
        osmium.SimpleHandler.__init__(self)

        self.strict_mode = True
        self.bounds = None
        self.POI = False

        self.osm_node_dict = {}
        self.osm_node_id_list = []
        self.osm_node_coord_list = []

        self.osm_way_dict = {}
        self.relation_list = []

    def node(self, n):
        osm_node_id = str(n.id)
        lon, lat = n.location.lon, n.location.lat
        node_geometry = geometry.Point(lon, lat)
        in_region = False if self.strict_mode and (not node_geometry.within(self.bounds)) else True

        osm_node_name = n.tags.get('name')
        osm_highway = n.tags.get('highway')
        ctrl_type = 'signal' if (osm_highway is not None) and 'signal' in osm_highway else None

        node = OSMNode(osm_node_name, osm_node_id, node_geometry, in_region, osm_highway, ctrl_type)
        self.osm_node_dict[node.osm_node_id] = node

        self.osm_node_id_list.append(osm_node_id)
        self.osm_node_coord_list.append((lon, lat))
        del n

    def way(self, w):
        way = Way()
        way.osm_way_id = str(w.id)
        way.ref_node_id_list = [str(node.ref) for node in w.nodes]

        way.highway = w.tags.get('highway')
        way.railway = w.tags.get('railway')
        way.aeroway = w.tags.get('aeroway')

        lane_info = w.tags.get('lanes')
        if lane_info is not None:
            lanes = re.findall(r'\d+\.?\d*', lane_info)
            if len(lanes) > 0:
                way.lanes = int(float(lanes[0]))  # in case of decimals
            else:
                logger = getLogger()
                if logger: logger.warning(f'new lanes type detected at way {way.osm_way_id}, {lane_info}')
        lane_info = w.tags.get('lanes:forward')
        if lane_info is not None:
            try:
                way.forward_lanes = int(lane_info)
            except:
                pass
        lane_info = w.tags.get('lanes:backward')
        if lane_info is not None:
            try:
                way.backward_lanes = int(lane_info)
            except:
                pass

        way.turn_lanes = w.tags.get('turn:lanes')
        way.turn_lanes_forward = w.tags.get('turn:lanes:forward')
        way.turn_lanes_backward = w.tags.get('turn:lanes:backward')

        way.name = w.tags.get('name')

        maxspeed_info = w.tags.get('maxspeed')
        if maxspeed_info is not None:
            try:
                way.maxspeed = int(float(maxspeed_info))
            except ValueError:
                speeds = re.findall(r'\d+\.?\d* mph', maxspeed_info)
                if len(speeds) > 0:
                    way.maxspeed = int(float(speeds[0][:-4]) * 1.6)
                else:
                    speeds = re.findall(r'\d+\.?\d* km/h', maxspeed_info)
                    if len(speeds) > 0:
                        way.maxspeed = int(float(speeds[0][:-5]))
                    else:
                        logger = getLogger()
                        if logger: logger.warning(f'new maxspeed type detected at way {way.osm_way_id}, {maxspeed_info}')

        oneway_info = w.tags.get('oneway')
        if oneway_info is not None:
            if oneway_info == 'yes' or oneway_info == '1':
                way.oneway = True
            elif oneway_info == 'no' or oneway_info == '0':
                way.oneway = False
            elif oneway_info == '-1':
                way.oneway = True
                way.is_reversed = True
            elif oneway_info in ['reversible', 'alternating']:
                # todo: reversible, alternating: https://wiki.openstreetmap.org/wiki/Tag:oneway%3Dreversible
                way.oneway = False
            else:
                logger = getLogger()
                if logger: logger.warning(f'new lane type detected at way {way.osm_way_id}, {oneway_info}')

        way.junction = w.tags.get('junction')
        way.area = w.tags.get('area')
        way.motor_vehicle = w.tags.get('motor_vehicle')
        way.motorcar = w.tags.get('motorcar')
        way.service = w.tags.get('service')
        way.foot = w.tags.get('foot')
        way.bicycle = w.tags.get('bicycle')
        way.building = w.tags.get('building')
        way.amenity = w.tags.get('amenity')
        way.leisure = w.tags.get('leisure')

        self.osm_way_dict[way.osm_way_id] = way
        del w


    def relation(self, r):
        if not self.POI: return

        relation = Relation()
        relation.osm_relation_id = str(r.id)

        relation.building = r.tags.get('building')
        relation.amenity = r.tags.get('amenity')
        relation.leisure = r.tags.get('leisure')
        if (relation.building is None) and (relation.amenity is None):
            return
        relation.name = r.tags.get('name')

        for member in r.members:
            member_id, member_type, member_role = member.ref, member.type, member.role
            member_id_str = str(member_id)
            member_type_lc = member_type.lower()
            if member_type_lc == 'n':
                relation.member_id_list.append(member_id_str)
            elif member_type_lc == 'w':
                relation.member_id_list.append(member_id_str)
            elif member_type_lc == 'r':
                pass
            else:
                logger = getLogger()
                if logger: logger.warning(f'new member type at relation {relation.osm_relation_id}, {member_type}')
            relation.member_type_list.append(member_type_lc)
            relation.member_role_list.append(member_role)

        self.relation_list.append(relation)
        del r


def _processNodes(net, h):
    coord_array = np.array(h.osm_node_coord_list)
    central_lon, central_lat = np.mean(coord_array, axis=0)
    central_lon, central_lat = float(central_lon), float(central_lat)
    northern = True if central_lat >= 0 else False

    xs, ys = from_latlon(coord_array[:, 0], coord_array[:, 1], central_lon)
    for node_no, node_id in enumerate(h.osm_node_id_list):
        node = h.osm_node_dict[node_id]
        node.geometry_xy = geometry.Point((xs[node_no], ys[node_no]))

    net.osm_node_dict = h.osm_node_dict
    net.GT = GeoTransformer(central_lon, central_lat, northern)


def _processWays(net, h):
    for osm_way_id, osm_way in h.osm_way_dict.items():
        try:
            osm_way.ref_node_list = [net.osm_node_dict[ref_node_id] for ref_node_id in osm_way.ref_node_id_list]
            net.osm_way_dict[osm_way_id] = osm_way
        except KeyError as e:
            print(f'WARNING: ref node {e} in way {osm_way_id} is not defined, way {osm_way_id} will not be imported')


def _processRelations(net, h):
    for relation in h.relation_list:
        valid = True
        for member_no, member_id in enumerate(relation.member_id_list):
            member_type = relation.member_type_list[member_no]
            if member_type == 'n':
                try:
                    relation.member_list.append(net.osm_node_dict[member_id])
                except KeyError as e:
                    print(f'WARNING: ref node {e} in relation {relation.osm_relation_id} is not defined, relation {relation.osm_relation_id} will not be imported')
                    valid = False
                    break
            elif member_type == 'w':
                try:
                    relation.member_list.append(net.osm_way_dict[member_id])
                except KeyError as e:
                    print(f'WARNING: ref way {e} in relation {relation.osm_relation_id} is not defined, relation {relation.osm_relation_id} will not be imported')
                    valid = False
                    break
            else:
                pass

        if valid:
            net.osm_relation_list.append(relation)


def _getBounds(filename, bbox):
    """
    Get network bounds from osm file or user given bbox

    Parameters
    ----------
    filename: str
        filename of the osm file
    bbox: tuple
        user given bound box

    Returns
    -------
    bounds: geometry.Polygon
        network boundary
    """

    try:
        f = osmium.io.Reader(filename)
    except:
        sys.exit(f'Error: filepath {filename} contains invalid characters. Please use english characters only')

    if bbox is None:
        header = f.header()
        box = header.box()
        bottom_left = box.bottom_left
        top_right = box.top_right
        try:
            minlat, minlon = bottom_left.lat, bottom_left.lon
            maxlat, maxlon = top_right.lat, top_right.lon
        except:
            minlat, minlon = og_settings.default_bounds['minlat'], og_settings.default_bounds['minlon']
            maxlat, maxlon = og_settings.default_bounds['maxlat'], og_settings.default_bounds['maxlon']
    else:
        minlat, minlon, maxlat, maxlon = bbox

    bounds = geometry.Polygon([(minlon, maxlat), (maxlon, maxlat), (maxlon, minlat), (minlon, minlat)])
    return bounds


def readOSMFile(filename, POI, strict_mode, bbox):
    """
    Get nodes, ways and relations from osm file

    Parameters
    ----------
    filename: str
        filename of the osm file
    POI: bool
        get POI information or not
    strict_mode: bool
        use strcit_mode or not.
    bbox: tuple
        user given network boundary

    Returns
    -------
    osmnet: OSMNetwork
        OSMNetwork from the osm file
    """

    if og_settings.verbose:
        print('  reading osm file')

    osmnet = OSMNetwork()

    osmnet.bounds = _getBounds(filename, bbox)

    h = NWRHandler()
    h.strict_mode = strict_mode
    h.bounds = osmnet.bounds
    h.POI = POI
    h.apply_file(filename)

    _processNodes(osmnet,h)
    _processWays(osmnet,h)
    _processRelations(osmnet, h)

    return osmnet





